# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
"""Environment creation, import, deletion dialogs."""

from .backup_dialogs import *  # noqa: F403
from .clone_dialogs import *  # noqa: F403
from .conflict_dialogs import *  # noqa: F403
from .create_dialogs import *  # noqa: F403
from .import_dialogs import *  # noqa: F403
from .multiaccount_dialogs import *  # noqa: F403
from .remove_dialogs import *  # noqa: F403
